from .Iscca import iscca
from .Iscca_cv import iscca_cv
from .Iscca_plot import iscca_plot
from .Preview_cca import preview_cca
from .Meta_scca import meta_scca
from .Scca import scca
from .Ispca import ispca
from .Ispca_cv import ispca_cv
from .Ispca_plot import ispca_plot
from .Preview_pca import preview_pca
from .Meta_spca import meta_spca
from .Spca import spca
from .Ispls import ispls
from .Ispls_cv import ispls_cv
from .Ispls_plot import ispls_plot
from .Preview_pls import preview_pls
from .Meta_spls import meta_spls
from .Spls import spls

__all__ = ['iscca', 'iscca_cv', 'iscca_plot', 'preview_cca','meta_scca','scca', 
           'ispca','ispca_cv','ispca_plot','preview_pca','meta_spca','spca',
           'ispls','ispls_cv','ispls_plot','preview_pls','meta_spls','spls']